/*
 * Wildcard TDM400P TDM FXS/FXO Interface Driver for DAHDI Telephony interface
 *
 * Written by Mark Spencer <markster@digium.com>
 *            Matthew Fredrickson <creslin@digium.com>
 *
 * Copyright (C) 2001-2008, Digium, Inc.
 *
 * All rights reserved.
 *
 */

/*
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2 as published by the
 * Free Software Foundation. See the LICENSE file included with
 * this program for more details.
 */
#include <linux/clk.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/ctype.h>
#include <linux/kmod.h>
#include <linux/moduleparam.h>
#include <linux/sched.h>
#include <linux/list.h>
#include <linux/delay.h>
#include <linux/mutex.h>
#include <linux/gpio.h>
#include <linux/dmw96dma.h>
#include <linux/err.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/ratelimit.h>
#include <linux/dmw96dma.h>
#include <sound/dmw-tdm.h>

#include <mach/hardware.h>
#include <mach/irqs.h>
#include <mach/reset.h>

#include "proslic.h"
#include <dahdi/kernel.h>
#include <dahdi/wctdm_user.h>
#include "fxo_modes.h"


#include <linux/spi/spi.h>
#include <linux/spi/spidev.h>
#include <linux/spi/spi_dmw96.h>

#include <sound/si3050.h>
#include <linux/kthread.h>
#include "dmw96_tdm.h"

#define USE_MINI_SPI
#if defined( USE_MINI_SPI )
#include <linux/spi/mini.h>
#endif

//=================
// Defines
//=================
#define MAGIC_INVALID	0x20121022

//=================
// Typedefs
//=================


//=================
// Functions
//=================
//#define AUDIO_RINGCHECK
//#define USE_DUMMY_FIFO_BANK

#define DEFAULT_TDM_BUS			(2)
#define DEFAULT_SAMPLE_RATE			(8000)
#define DEFAULT_SAMPLE_BITS 			(8)
#define DEFAULT_TIME_SLOTS 			(32)

/* SPI: FXO daisy chain bit reversed order for chipts 1, 2, 3, and 4 on a chain */
#define FXO_WR_CMD_CARD(card)	(0x20|(card==0?0:card==1?0x08:card==2?0x04:card==3?0x0c:0))
#define FXO_RD_CMD_CARD(card)	(0x60|(card==0?0:card==1?0x08:card==2?0x04:card==3?0x0c:0))


#define NUM_FXO_REGS 60

#define WC_MAX_IFACES 128

#define DEFAULT_RING_DEBOUNCE	256		/* Ringer Debounce (256 ms) */

#define POLARITY_DEBOUNCE 	64		/* Polarity debounce (64 ms) */

#define OHT_TIMER		6000	/* How long after RING to retain OHT */

/* NEON MWI pulse width - Make larger for longer period time
 * For more information on NEON MWI generation using the proslic
 * refer to Silicon Labs App Note "AN33-SI321X NEON FLASHING"
 * RNGY = RNGY 1/2 * Period * 8000
 */
#define NEON_MWI_RNGY_PULSEWIDTH	0x3e8	/*=> period of 250 mS */

#define NUM_CARDS 4

#define MAX_ALARMS 10

#define MOD_TYPE_FXS	0
#define MOD_TYPE_FXO	1

#define MINPEGTIME	10 * 8		/* 30 ms peak to peak gets us no more than 100 Hz */
#define PEGTIME		50 * 8		/* 50ms peak to peak gets us rings of 10 Hz or more */
#define PEGCOUNT	5		/* 5 cycles of pegging means RING */

#define NUM_CAL_REGS 12

//=================
// Typedefs
//=================

/* DMA channel states */
typedef enum {

        /* Not running */
        kDMA_ST_IDLE,

        /* Running 'Expicitly', due to explicit ALSA _prepare() callback */
        kDMA_ST_EXPLCT_RUNNING,

        /* Running 'Impicitly', due to _prepar() of Other(!) Direction stream.
         * Note: Must run dma on other direction as well, due to dmw96's
         *       TDM machine HW restrictions.
         */
        kDMA_ST_IMPLCT_RUNNING,

} eDMA_ST;

#define MAX_DMA_BUF	16
struct tdm {
	unsigned int index;
        void *reg_base;
        struct clk *tdm_clk;
        int running;

        void *tx_buffer[MAX_DMA_BUF];
        void *rx_buffer[MAX_DMA_BUF];
	int tx_buffer_ping_pong;
	int rx_buffer_ping_pong;
        u32 *rx_ptr;
        u32 *tx_ptr;
	int xfering_pcm;

	eDMA_ST tx_dma_state;
	eDMA_ST rx_dma_state;
	int tdm_buffer_idx_dma;
	int tdm_buffer_idx_sw;
        struct dmw96dma_list *tx_dma_list;
        struct dmw96dma_list *rx_dma_list;
	struct tasklet_struct iotask;
	unsigned int usecount;
	unsigned int reg_size;
	unsigned int irqno;
	unsigned int slots;
	unsigned int srate;
	unsigned int sbits;
	unsigned int master;
	const char *clk;
	unsigned int dma_rx;
	unsigned int dma_tx;
	unsigned int rx_size;
	unsigned int tx_size;
	unsigned int rx_edge;
	unsigned int tx_edge;
	unsigned int reset;
	unsigned int chunks;
	unsigned int dma_enable;
	struct timer_list timer;
	unsigned int dma_cancel;
	unsigned int offset;
};

struct calregs {
	unsigned char vals[NUM_CAL_REGS];
};

enum proslic_power_warn {
	PROSLIC_POWER_UNKNOWN = 0,
	PROSLIC_POWER_ON,
	PROSLIC_POWER_WARNED,
};

enum battery_state {
	BATTERY_UNKNOWN = 0,
	BATTERY_PRESENT,
	BATTERY_LOST,
};

struct wctdm {
	struct spi_device *dev;
	struct snd_soc_codec codec;
	struct si3050_platform_data *pdata;
	char *variety;
	struct dahdi_span span;
	struct dahdi_device *ddev;
	unsigned char ios;
	int usecount;
	unsigned int intcount;
	int dead;
	int iface_index;
	int pos;
	int flags[NUM_CARDS];
	int freeregion;
	int alt;
	int curcard;
	int cardflag;		/* Bit-map of present cards */
	enum proslic_power_warn proslic_power;
	spinlock_t lock;

	union {
		struct fxo {
#ifdef AUDIO_RINGCHECK
			unsigned int pegtimer;
			int pegcount;
			int peg;
			int ring;
#endif			
			int wasringing;
			int lastrdtx;
			
			int ringdebounce;
			int offhook;
			int battdebounce;
			int battalarm;
			enum battery_state battery;
		        int lastpol;
		        int polarity;
		        int polaritydebounce;
			long long ringdebounce_last;
			long long battdebounce_last;
			long long battalarm_last;
			long long polaritydebounce_last;
			int ringtimeout;
		} fxo;
		struct fxs {
			int oldrxhook;
			int debouncehook;
			int lastrxhook;
			int debounce;
			int ohttimer;
			int idletxhookstate;		/* IDLE changing hook state */
			int lasttxhook;
			int palarms;
			int reversepolarity;		/* Reverse Line */
			int mwisendtype;
			struct dahdi_vmwi_info vmwisetting;
			int vmwi_active_messages;
			u32 vmwi_lrev:1; /*MWI Line Reversal*/
			u32 vmwi_hvdc:1; /*MWI High Voltage DC Idle line*/
			u32 vmwi_hvac:1; /*MWI Neon High Voltage AC Idle line*/
			u32 neonringing:1; /*Ring Generator is set for NEON*/
			struct calregs calregs;
		} fxs;
	} mod[NUM_CARDS];

	/* Receive hook state and debouncing */
	int modtype[NUM_CARDS];
	unsigned char reg0shadow[NUM_CARDS];
	unsigned char reg1shadow[NUM_CARDS];

	unsigned long ioaddr;
	dma_addr_t 	readdma;
	dma_addr_t	writedma;
	volatile unsigned int *writechunk;				/* Double-word aligned write memory */
	volatile unsigned int *readchunk;				/* Double-word aligned read memory */
	struct dahdi_chan _chans[NUM_CARDS];
	struct dahdi_chan *chans[NUM_CARDS];
	int irq;
	unsigned int num_cards;
	unsigned int check;
};


struct wctdm_desc {
	char *name;
	int flags;
};

static struct task_struct *check_hook_task;
//=================
// Functions
//=================
/*
 * Public Functions
 */

/*
 * Private Functions
 */
static void wctdm_release(struct wctdm *wc);
static void gs_check_hook(void);
static inline void wctdm_voicedaa_check_hook(struct wctdm *wc, int card);
#if defined( USE_MINI_SPI )
static void gs_read_hook_regs(void);
#endif

/*
 * Public Data
 */


/*
 * Private Data
 */
static struct wctdm *ifaces[WC_MAX_IFACES];
static int ifaces_num = 0;

struct tdm *tdm = NULL;
static struct wctdm_desc gs_gxe51xx_desc = { "GS FXO PSTN BANK for GXE51XX", 0 };

static unsigned int disable_led;
static int ringtimeout = 12;
static unsigned int fxo_acim;
static unsigned int fxo_override_acim;
static unsigned int check_prio = MAX_RT_PRIO-8;
static unsigned int check_tv = 1;
static unsigned int tdm_bus = DEFAULT_TDM_BUS;
static unsigned int tdm_slots = DEFAULT_TIME_SLOTS;
static unsigned int tdm_srate = DEFAULT_SAMPLE_RATE;
static unsigned int tdm_sbits = DEFAULT_SAMPLE_BITS;
static unsigned int tdm_rxedge = 1;
static unsigned int tdm_txedge = 0;
static unsigned int tdm_master = 1;
static unsigned int tdm_dma = 1;
static unsigned int tdm_dma_chunks = DAHDI_CB_SIZE;
/* module parameters */
static unsigned int fxovoltage;
static unsigned int battdebounce;
static unsigned int battalarm;
static unsigned int battthresh;
static int ringdebounce = DEFAULT_RING_DEBOUNCE;
/* times 4, because must be a multiple of 4ms: */
static int dialdebounce = 8 * 8;
static int fwringdetect = 0;
static int debug = 1;
static int robust = 0;
static int timingonly = 0;
static int lowpower = 0;
static int boostringer = 0;
static int fastringer = 0;
static int _opermode = 0;
static char *opermode = "FCC";
static int fxshonormode = 0;
static int alawoverride = 0;
static int fastpickup = 0;
static int fxotxgain = 0;
static int fxorxgain = 0;
static int fxstxgain = 0;
static int fxsrxgain = 0;


//unsigned long long dp_clk_freq;
static s16 sine[] = {
             0,
          6269,
         11584,
         15136,
         16383,
         15137,
         11586,
          6271,
             1,
         -6268,
        -11583,
        -15136,
        -16383,
        -15137,
        -11587,
         -6272
};

/*
 *  Define for audio vs. register based ring detection
 *  
 */
/* #define AUDIO_RINGCHECK  */

/*
  Experimental max loop current limit for the proslic
  Loop current limit is from 20 mA to 41 mA in steps of 3
  (according to datasheet)
  So set the value below to:
  0x00 : 20mA (default)
  0x01 : 23mA
  0x02 : 26mA
  0x03 : 29mA
  0x04 : 32mA
  0x05 : 35mA
  0x06 : 37mA
  0x07 : 41mA
*/
static int loopcurrent = 20;
#define POLARITY_XOR (\
		(reversepolarity != 0) ^ (fxs->reversepolarity != 0) ^\
		(fxs->vmwi_lrev != 0) ^\
		((fxs->vmwisetting.vmwi_type & DAHDI_VMWI_HVAC) != 0))

static int reversepolarity = 0;

#ifdef AUDIO_RINGCHECK
static inline void ring_check(struct wctdm *wc, int card)
{
	int x;
	short sample;
	if (wc->modtype[card] != MOD_TYPE_FXO)
		return;
	wc->mod[card].fxo.pegtimer += DAHDI_CHUNKSIZE;
	for (x=0;x<DAHDI_CHUNKSIZE;x++) {
		/* Look for pegging to indicate ringing */
		sample = DAHDI_XLAW(wc->chans[card]->readchunk[x], (wc->chans[card]));
		if ((sample > 10000) && (wc->mod[card].fxo.peg != 1)) {
			if (debug > 1) printk(KERN_DEBUG "High peg!\n");
			if ((wc->mod[card].fxo.pegtimer < PEGTIME) && (wc->mod[card].fxo.pegtimer > MINPEGTIME))
				wc->mod[card].fxo.pegcount++;
			wc->mod[card].fxo.pegtimer = 0;
			wc->mod[card].fxo.peg = 1;
		} else if ((sample < -10000) && (wc->mod[card].fxo.peg != -1)) {
			if (debug > 1) printk(KERN_DEBUG "Low peg!\n");
			if ((wc->mod[card].fxo.pegtimer < (PEGTIME >> 2)) && (wc->mod[card].fxo.pegtimer > (MINPEGTIME >> 2)))
				wc->mod[card].fxo.pegcount++;
			wc->mod[card].fxo.pegtimer = 0;
			wc->mod[card].fxo.peg = -1;
		}
	}
	if (wc->mod[card].fxo.pegtimer > PEGTIME) {
		/* Reset pegcount if our timer expires */
		wc->mod[card].fxo.pegcount = 0;
	}
	/* Decrement debouncer if appropriate */
	if (wc->mod[card].fxo.ringdebounce)
		wc->mod[card].fxo.ringdebounce--;
	if (!wc->mod[card].fxo.offhook && !wc->mod[card].fxo.ringdebounce) {
		if (!wc->mod[card].fxo.ring && (wc->mod[card].fxo.pegcount > PEGCOUNT)) {
			/* It's ringing */
			if (debug)
				printk(KERN_DEBUG "RING on FXO%d!\n", wc->chans[card].channo);
			if (!wc->mod[card].fxo.offhook)
				dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_RING);
			wc->mod[card].fxo.ring = 1;
		}
		if (wc->mod[card].fxo.ring && !wc->mod[card].fxo.pegcount) {
			/* No more ring */
			if (debug)
				printk(KERN_DEBUG "NO RING on FXO%d!\n", wc->chans[card].channo);
			dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
			wc->mod[card].fxo.ring = 0;
		}
	}
}
#endif

/*
 * TDM Controls 
 */

static unsigned long tdm_read(struct tdm *tdm, int reg)
{
        return readl(tdm->reg_base + reg);
}

static void tdm_write(struct tdm *tdm, int reg, unsigned long value)
{
//printk("tdm_write reg = 0x%04x, val = 0x%08x\n", reg, value);
        writel(value, tdm->reg_base + reg);
}

static void tdm_clk_disable(struct tdm *tdm)
{
        if (tdm->running) {
                if(tdm->master) clk_disable(tdm->tdm_clk);
                tdm->running = 0;
        }
}

static void tdm_clk_enable(struct tdm *tdm)
{
        if (!tdm->running) {
                if(tdm->master) clk_enable(tdm->tdm_clk);
                tdm->running = 1;
        }
}

static void tdm_set_reset(struct tdm *tdm, int enable)
{
        if (enable)
                reset_set(tdm->reset);
        else
                reset_release(tdm->reset);
}

static void tdm_reset(struct tdm *tdm)
{
	tdm_write(tdm, TDM_ENABLE, 0);

        /* TDM clock mus be disabled when doing the ARM TDM SW-Reset. */
        tdm_clk_disable(tdm);

        /* Reset TDM register block */
        tdm_set_reset(tdm, 1);
        udelay(10);
        tdm_set_reset(tdm, 0);

        /* Enable TDM clock - for safe configuration of the HW machine internal
         * states */
        tdm_clk_enable(tdm);
}

unsigned int slots2banksize(unsigned int slots)
{
	switch(slots)
	{
	case 2:
		return 0;
	case 4:
		return 1;
	case 8:
		return 2;
	case 16:
		return 3;
	case 32:
		return 4;
	default:
		return 4;
	}
}

static unsigned int slots2banks(unsigned int slots)
{
	switch(slots)
	{
	case 2:
		return 64;
	case 4:
		return 32;
	case 8:
		return 16;
	case 16:
		return 8;
	case 32:
		return 4;
	default:
		return 4;
	}
}

void tdm_setup(struct tdm *tdm)
{
	unsigned int val;
	long freq;
	int i,j,k;
	unsigned int tx_fifo_wm_lvl, rx_fifo_wm_lvl;
	unsigned int clk = tdm->slots*tdm->srate*tdm->sbits;
	unsigned int banksize;
	unsigned int banks;

	if(tdm->master)
	{
		freq = clk_round_rate(tdm->tdm_clk, clk );
		if (freq < 0) {
				printk("clk_round_rate(tdm, %d): %ld\n", clk, freq);
				return;
		}
		if (clk_set_rate(tdm->tdm_clk, freq)) {
				printk("clk_set_rate(tdm) failed!\n");
				return;
		}
		printk("TDM SCLK = %ldHz\n", freq);
	}

	/*
	 * We should sw-reset the ARM TDM block perior to any TDM configuration
	 * change.
	 */
	tdm_reset(tdm);

	/* DATA_WIDTH = 16, CS_D = 0 (audio/voice data time slot), TX_EN = 1,
	 * RX_EN = 1 */	
	k = -1;
	for (i=0; i<tdm->slots; i++)
	{
		k = i;
		val =  ((k << 8)|      /* ENTRY */
				(1 << 7) |  /* RX_EN = 1*/
				(1 << 6) |  /* TX_EN = 1*/
				(0 << 5) |      /* CS_D = 0 */
				(tdm->sbits & 0x1F));
		tdm_write(tdm, FDA_REG_ADDR(i), val);
	}

	/* Configure TDM as master SYNC/TX/RX Edges, 2 time-slots, for
	 * "Left-Justified" TDM protocol flavour
	 */
	val = 0x0000;
	if(!tdm->master)
	{
		val |= TDM_CONFIG_1_AUTO_SYNC;	
	}
	else
	{
		val |= TDM_CONFIG_1_FSYNC_MSTR_SLV;
		val |= TDM_CONFIG_1_CLK_MSTR_SLV;
	}
	val &= ~TDM_CONFIG_1_SYNC_EDGE;
	if(tdm->rx_edge) val |= TDM_CONFIG_1_RX_EDGE;
	if(tdm->tx_edge) val |= TDM_CONFIG_1_TX_EDGE;
	val |= ((tdm->slots-1) & TDM_CONFIG_1_TIME_SLOTS); /*: two time slots */

	tdm_write(tdm, TDM_CONFIG_1, val );

	banksize = slots2banksize(tdm->slots);
	banks = slots2banks(tdm->slots);
	/* Configure MSB first, 64 banks, 2 fifo lines */
	tdm_write(tdm, TDM_CONFIG_2,
		TDM_CONFIG_2_SHIFT |  /* MSB transmit first */
		TDM_CONFIG_2_TRUNCATE|banksize);

	/* Configure Fsync signal parameters, according to TDM Format: */
	val = 0x0000;
	if(!tdm->master)
	{
		val |= (7 << 8);
	}

	tdm_write(tdm, TDM_FSYNC, val);

	/* Configure Fsync High for 1 sclk cycles (16 bits) */
	tdm_write(tdm, TDM_FSYNC_DUR, 0);

	/* Configure TX & RX FIFOs Water-Mark */
	val = 0x0000;
	if(tdm->dma_enable)
	{
		tx_fifo_wm_lvl = (banks>>1)-1;
		rx_fifo_wm_lvl = (banks>>1)-1;
	}
	else
	{
		tx_fifo_wm_lvl = 0;
		rx_fifo_wm_lvl = 0;
	}

	val |= (tx_fifo_wm_lvl & 0x003F) << 0;
	val |= (rx_fifo_wm_lvl & 0x003F) << 8;    /* bits 13:8 */

	tdm_write(tdm, TDM_FIFO_WM, val);

	/* Fill-in Tx fifo */
	j = tx_fifo_wm_lvl + 1;
	while (j--)
	{
		for ( i = 0; i < tdm->slots; i++ )
		{
			tdm_write(tdm, TDM_TX_FIFO, 0x0000);
		}
	}

	/* Empty Rx fifo */
	j = rx_fifo_wm_lvl + 1;
	while (j--)
	{
		for (i=0; i<tdm->slots; i++)
		{
			tdm_read(tdm, TDM_RX_FIFO);
		}
	}
}

static inline void gstdm_transmitprep(struct tdm *tdm)
{
	int x;
	volatile unsigned int *writechunk;
	struct wctdm *wc = NULL;
	int i;
	unsigned int card;
	unsigned int k;
	unsigned int j;
	unsigned char *echunk;

	writechunk = tdm->tx_buffer[tdm->tx_buffer_ping_pong];
	for(j=0; j<tdm->chunks; j++)
	{
		for(i=0;(wc = ifaces[i]);i++)
		{
			/* Calculate Transmission */
			_dahdi_transmit(&wc->span);
		}

		for(i=0;(wc = ifaces[i]);i++)
		{
			for (card=0;card<wc->num_cards;card++) {
				if (!(wc->cardflag & (1 << card))) continue;
				x = j*tdm->slots*DAHDI_CHUNKSIZE+i*NUM_CARDS+card+tdm->offset;
				for(k=0; k<DAHDI_CHUNKSIZE; k++)
				{
					writechunk[x]  = (wc->chans[card]->writechunk[k])<<24;
					x+=tdm->slots;
				}
#if 0
				/* copy writechunk for echo cancel */
				echunk = wc->chans[card]->writechunk+(tdm->chunks-j-1)*DAHDI_CHUNKSIZE;
				if(echunk!=wc->chans[card]->writechunk)
				{
					memcpy(echunk, wc->chans[card]->writechunk, DAHDI_CHUNKSIZE);
				}
#endif
			}
		}
	}
}

static inline void gstdm_receiveprep(struct tdm *tdm)
{
	int x;
	volatile unsigned int *readchunk;
	struct wctdm *wc = NULL;
	int i;
	unsigned int card;
	unsigned int k;
	unsigned int j;
	unsigned char *echunk;

	readchunk = tdm->rx_buffer[tdm->rx_buffer_ping_pong];
	for(j=0; j<tdm->chunks; j++)
	{
		for(i=0;(wc = ifaces[i]);i++)
		{
			for (card=0;card<wc->num_cards;card++) {
				if (!(wc->cardflag & (1 << card))) continue;
				x = j*tdm->slots*DAHDI_CHUNKSIZE+i*NUM_CARDS+card+tdm->offset;
				for(k=0; k<DAHDI_CHUNKSIZE; k++)
				{
					wc->chans[card]->readchunk[k] = readchunk[x] >> 24;
					x+=tdm->slots;
				}
#if 0
				echunk = wc->chans[card]->writechunk+(tdm->chunks-j-1)*DAHDI_CHUNKSIZE;
				dahdi_ec_chunk(wc->chans[card], wc->chans[card]->readchunk, echunk);
#endif
			}
		}

		for(i=0;(wc = ifaces[i]);i++)
		{
			_dahdi_receive(&wc->span);
		}
	}
}

static void gstdm_led_timer(unsigned long data)
{
	struct tdm *tdm;
	struct si3050_platform_data *pdata;
	struct wctdm *wc = NULL;
	int i;
	unsigned int card;
	struct fxo *fxo;
	int led_gpio;
	static unsigned int count = 0;
	int value;

	count++;
	tdm = (struct tdm *)data;
	if(unlikely(disable_led))
	{
		mod_timer(&tdm->timer, jiffies+msecs_to_jiffies(500));
		return;
	}
	for(i=0;(wc = ifaces[i]);i++)
	{
		pdata = wc->pdata;
		if(!pdata) continue;
		for (card=0;card<wc->num_cards;card++) {
			if((!pdata->led_gpio) || (card>=pdata->led_count)) continue;

			led_gpio = pdata->led_gpio[card];
			if(!led_gpio) continue;

			if (!(wc->cardflag & (1 << card)))
			{
				/* unusable, led on 2 second off 2 second */
				gpio_set_value(led_gpio, count&0x4);
				continue;
			}

			fxo = &wc->mod[card].fxo;

			if(BATTERY_PRESENT==fxo->battery)
			{
				if(fxo->offhook)
				{
					/* incall, led on 0.5 second off 0.5 second */
					value = count&0x1;
				}
				else if(fxo->ringtimeout>0)
				{
					/* ring, led on 1 second off 1 second */
					value = count&0x2;
					fxo->ringtimeout--;
				}
				else
				{
					/* connected, led on */
					value = 1;
				}
			}
			else
			{
				/* not connected, led off */
				value = 0;
			}
			gpio_set_value(led_gpio, value);
		}
	}
	mod_timer(&tdm->timer, jiffies+msecs_to_jiffies(500));
}

static void rx_dma_buf_done(int status, void *context)
{
	struct tdm *tdm = (struct tdm*)context;

	if (unlikely(status))
	{
		if(!tdm->dma_cancel) printk_ratelimited("TDM%d dma rx err, %d\n", tdm->index, status);
	}
	else
	{
		if(tdm->usecount) gstdm_receiveprep(tdm);
	}
	tdm->rx_buffer_ping_pong++;
	tdm->rx_buffer_ping_pong %= MAX_DMA_BUF;
	return;
}

static void tx_dma_buf_done(int status, void *context)
{
	struct tdm *tdm = (struct tdm*)context;

	if (unlikely(status))
	{
		if(!tdm->dma_cancel) printk_ratelimited("TDM%d dma tx err, %d\n", tdm->index, status);
	}
	else
	{
		if(tdm->usecount) gstdm_transmitprep(tdm);
	}

	tdm->tx_buffer_ping_pong++;
	tdm->tx_buffer_ping_pong %= MAX_DMA_BUF;

	return;
}

static void finish_out(int status, void *context)
{
        struct tdm *tdm = (struct tdm *)context;
        int ret;

	if(tdm->dma_cancel) return;
	if(status)
	{
		printk_ratelimited("TDM%d tx fail, %d\n", tdm->index, status);
	}
	
        if ((ret = dmw96dma_submit(tdm->tx_dma_list)))
                printk_ratelimited("TDM%d submit tx list fail, %d\n", tdm->index, ret);
}

static void finish_in(int status, void *context)
{
        struct tdm *tdm = (struct tdm *)context;
        int ret;
	
	if(tdm->dma_cancel) return;
	if(status)
	{
		printk_ratelimited("TDM%d rx fail, %d\n", tdm->index, status);
	}
        if ((ret = dmw96dma_submit(tdm->rx_dma_list)))
                printk_ratelimited("TDM%d submit rx list fail, %d\n", tdm->index, ret);
}

static void tdm_free_dma_data(struct tdm *tdm)
{
	unsigned int i;

	tdm->dma_cancel = 1;
	if (tdm->rx_dma_list) dmw96dma_cancel(tdm->rx_dma_list);
	if (tdm->tx_dma_list) dmw96dma_cancel(tdm->tx_dma_list);

	/* Free TX DTM buffers */
	for(i=0; i< MAX_DMA_BUF; i++)
	{
		if (tdm->tx_buffer[i]) {
			kfree(tdm->tx_buffer[i]);
		}
	}

        if (tdm->tx_dma_list) {
                dmw96dma_free_list(tdm->tx_dma_list);
        }

	/* Free RX DTM buffers */
	for(i=0; i< MAX_DMA_BUF; i++)
	{
		if (tdm->rx_buffer[i]) {
			kfree(tdm->rx_buffer[i]);
		}
	}

        if (tdm->rx_dma_list) {
                dmw96dma_free_list(tdm->rx_dma_list);
        }
}

int fill_txbuf(struct tdm *tdm)
{
	u32 *buf;
	unsigned int i;
	unsigned int j;
	unsigned int off;
	unsigned int count;
	unsigned int index;

	count = (tdm->tx_size>>2)/tdm->slots;
	index = 0;
	for(i=0; i< MAX_DMA_BUF; i++)
	{
		/* fill with pattern */
		buf = tdm->tx_buffer[i];
		for (off = 0; off<count ;off++)
		{
			for (j=0; j<tdm->slots; j++)
			{
			   *buf++ = ((u32)(sine[index&0xf])) << 16;
			}
			index++;
		}
	}

	return 0;
}

int tdm_start_output(struct tdm *tdm)
{
	int ret = 0, i;
	int int_en = 0;
	unsigned int tx_size;
	unsigned int rx_size;

        /* Pre-allocate DMA buffers : */
        tdm->tx_dma_state = kDMA_ST_IDLE;
        tdm->rx_dma_state = kDMA_ST_IDLE;

	if(tdm_dma)
	{
		if(tdm_dma_chunks>DAHDI_CB_SIZE)
		{
			tdm_dma_chunks = DAHDI_CB_SIZE;
		}
		else
		{
			if(tdm_dma_chunks<1) tdm_dma_chunks = 1;
		}
		tdm->chunks = tdm_dma_chunks;
	}
	else tdm->chunks = 1;
	tx_size = 4*tdm->slots*DAHDI_MAX_CHUNKSIZE*tdm->chunks;
	rx_size = 4*tdm->slots*DAHDI_MAX_CHUNKSIZE*tdm->chunks;
	tdm->tx_size = tx_size;
	tdm->rx_size = rx_size;

	printk("MAX_DMA_BUF %d\n", MAX_DMA_BUF);
	for(i=0; i< MAX_DMA_BUF; i++)
	{
		tdm->tx_buffer[i] = kzalloc(tx_size, GFP_KERNEL | GFP_DMA );
		if (!tdm->tx_buffer[i])
			goto free_all;
	}

	for(i=0; i< MAX_DMA_BUF; i++)
	{
		tdm->rx_buffer[i] = kzalloc(rx_size, GFP_KERNEL | GFP_DMA );
		if (!tdm->rx_buffer[i])
			goto free_all;
	}

	tdm->tx_ptr = tdm->tx_buffer[0];
	tdm->rx_ptr = tdm->rx_buffer[0];

	fill_txbuf(tdm);

        tdm->tx_dma_list = dmw96dma_alloc_io_list(1, tdm->dma_tx, &finish_out, tdm);
        if (!tdm->tx_dma_list) {
                ret = -ENOMEM;
                goto free_all;
        }

	for(i=0; i< MAX_DMA_BUF; i++)
	{
		if ((ret = dmw96dma_add_io_transfer_buf(tdm->tx_dma_list, tdm->tx_buffer[i], tx_size, tx_dma_buf_done, tdm)))
			goto free_all;
	}

        tdm->rx_dma_list = dmw96dma_alloc_io_list(1, tdm->dma_rx, &finish_in, tdm);
        if (!tdm->rx_dma_list) {
                ret = -ENOMEM;
                goto free_all;
        }

	for(i=0; i< MAX_DMA_BUF; i++)
	{
		if ((ret = dmw96dma_add_io_transfer_buf(tdm->rx_dma_list, tdm->rx_buffer[i], rx_size, rx_dma_buf_done, tdm)))
			goto free_all;
	}

	if(tdm->dma_enable)
	{
		if ((ret = dmw96dma_submit(tdm->tx_dma_list)))
				goto free_all;
		int_en |= TDM_INT_EN_TX_EMPTY;
	}
	else
	{
		int_en |= TDM_INT_EN_TX_WM;
	}

	if(tdm->dma_enable)
	{
		if ((ret = dmw96dma_submit(tdm->rx_dma_list)))
			goto free_all;
		int_en |= TDM_INT_EN_RX_FULL;
	}
	else
	{
		int_en |= TDM_INT_EN_RX_WM;
	}

	tdm_write(tdm, TDM_INT_EN, int_en);

	tdm->xfering_pcm = 1;
	return 0;

free_all:
	tdm_free_dma_data(tdm);
        return ret;
}

static irqreturn_t tdm_interrupt_handler(int irq, void *data)
{
	struct tdm *tdm = (struct tdm *)data;
	int pending,i;
	u32 *ptr;
	u32 *end;

	pending = tdm_read(tdm, TDM_INT_STATUS) & tdm_read(tdm, TDM_INT_EN);
	//printk("TDM INT: 0x%08x\n", pending);
	if( pending & (TDM_INT_STATUS_RX_OVRN | TDM_INT_STATUS_TX_UNDRN))
	{
		printk_ratelimited("TDM%d overrun: 0x%08x\n", tdm->index, pending);
		tdm_write(tdm, TDM_INT_CLR, TDM_INT_CLR_RX_OVRN | TDM_INT_CLR_TX_UNDRN);
	}
	else
	{
		if(pending&TDM_INT_STATUS_FSYNC) printk_ratelimited("TDM%d unexpected FSYNC or no FSYNC\n", tdm->index);
		tdm_write(tdm, TDM_INT_CLR, 0xFFFFF);
	}


	if (pending & TDM_INT_STATUS_RX)
	{ /* RX */
		ptr = tdm->rx_ptr;;
		end = tdm->rx_buffer[tdm->rx_buffer_ping_pong]+tdm->rx_size;
		while ((tdm_read(tdm, TDM_STATUS) >> 8) & 0x7f)
		{
			for (i=0; i<tdm->slots; i++)
			{
				*ptr = tdm_read(tdm, TDM_RX_FIFO);
				ptr++;
				if(ptr >= end)
				{
					if(tdm->usecount) gstdm_receiveprep(tdm);
					tdm->rx_buffer_ping_pong++;
					tdm->rx_buffer_ping_pong%=MAX_DMA_BUF;
					ptr = tdm->rx_buffer[tdm->rx_buffer_ping_pong];
					end = (u32 *)(((char *)ptr)+tdm->rx_size);
				}
			}
			tdm->rx_ptr = ptr;
		}
	}

	if (pending & TDM_INT_STATUS_TX)
	{ /* TX */
		ptr = tdm->tx_ptr;
		end = tdm->tx_buffer[tdm->tx_buffer_ping_pong]+tdm->tx_size;
		while (tdm_read(tdm, TDM_STATUS) & 0x7f)
		{
			for (i=0; i<tdm->slots; i++)
			{
				tdm_write(tdm, TDM_TX_FIFO, *ptr);
				ptr++;
				if(ptr >= end)
				{
					tdm->tx_buffer_ping_pong++;
					tdm->tx_buffer_ping_pong%=MAX_DMA_BUF;
					if(tdm->usecount) gstdm_transmitprep(tdm);
					ptr = tdm->tx_buffer[tdm->tx_buffer_ping_pong];
					end = (u32 *)(((char *)ptr)+tdm->tx_size);
				}
			}
			tdm->tx_ptr = ptr;
		}
	}

	return IRQ_HANDLED;
}

static tdm_resource_t tdm_resource[3]=
{
		{DMW_TDM1_BASE, 0x2000, "dmw-tdm.0", DMW_IID_TDM1, 8, 9, RESET_TDM1},
		{DMW_TDM2_BASE, 0x2000, "dmw-tdm.1", DMW_IID_TDM2, 10, 11, RESET_TDM2},
		{DMW_TDM3_BASE, 0x2000, "dmw-tdm.2", DMW_IID_TDM3, 12, 13, RESET_TDM3},
};

int tdm_init(unsigned int index)
{
        int ret;

        tdm = kzalloc(sizeof(*tdm), GFP_KERNEL);
	memset(tdm, 0, sizeof(*tdm));

	tdm->tx_buffer_ping_pong = 0;
	tdm->rx_buffer_ping_pong = 0;

	index %= ARRAY_SIZE(tdm_resource);
	tdm->index = index;
	tdm->reg_size = tdm_resource[index].reg_size;
	tdm->clk = tdm_resource[index].clk;
	tdm->irqno = tdm_resource[index].irqno;
	tdm->dma_rx = tdm_resource[index].dma_rx;
	tdm->dma_tx = tdm_resource[index].dma_tx;
	tdm->slots = tdm_slots;
	tdm->srate = tdm_srate;
	tdm->sbits = tdm_sbits;
	tdm->master = tdm_master;
	tdm->offset = tdm_master?0:1;
	tdm->rx_edge = tdm_rxedge;
	tdm->tx_edge = tdm_txedge;
	tdm->reset = tdm_resource[index].reset;
	tdm->dma_enable = tdm_dma;
        tdm->reg_base = ioremap_nocache(tdm_resource[index].reg_base, tdm->reg_size);
        if (!tdm->reg_base) {
                printk("ioremap_nocache() failed for TDM%d\n", index);
                return -1;
        }

        tdm->tdm_clk = clk_get_sys(tdm->clk, NULL);

        if (IS_ERR(tdm->tdm_clk)) {
                ret = PTR_ERR(tdm->tdm_clk);
                printk("TDM%d clock get failed\n", index);
                goto err_ioremap;
        }

        tdm_setup(tdm);
        ret = request_irq(tdm->irqno, tdm_interrupt_handler, IRQF_DISABLED, tdm->clk, tdm);
        if (ret) {
                printk("TDM%d could not get irq: %d\n", index, ret);
                goto err_irq;
        }

        if ((ret = tdm_start_output(tdm))) {
                printk("tdm_start_output: failed with %d\n", ret);
                goto err_irq;
        }

        /* Enable TDM */
        tdm_write(tdm, TDM_ENABLE, tdm_read(tdm, TDM_ENABLE) | 1);

	setup_timer(&tdm->timer, gstdm_led_timer, (unsigned long)tdm);
	mod_timer(&tdm->timer, jiffies+msecs_to_jiffies(500));

        printk("TDM%d: %d slots, dma %d\n", index, tdm->slots, tdm->dma_enable);
        return 0;
err_irq:
        clk_put(tdm->tdm_clk);
err_ioremap:
        iounmap(tdm->reg_base);
        return ret;
}


void tdm_remove(void)
{
	if (tdm)
	{
		del_timer(&tdm->timer);
		free_irq(tdm->irqno, tdm);
		tdm_set_reset(tdm, 1);

		/* Free all allocated DMA memories and lists  */
		tdm_free_dma_data(tdm);
	        iounmap(tdm->reg_base);
		clk_put(tdm->tdm_clk);
	        kfree(tdm);
		tdm = NULL;
	}
	else
		printk("DAHDI TDM already freed\n");
}

static void wctdm_setreg(struct wctdm *wc, int card, unsigned char reg, unsigned char value)
{
#if defined( USE_MINI_SPI )
    unsigned char wrbuf[3];
    unsigned char val;
    int i;
    int chip_select;

#if defined( MINI_SPI_USE_MUTEX )
    mutex_lock( &mini_spi_mutex );
#else
    unsigned long flags;
    spin_lock_irqsave( &mini_spi_dat.lock, flags );
#endif

    chip_select = wc->dev->chip_select;

    wrbuf[0] = FXO_WR_CMD_CARD(card);
    wrbuf[1] = reg & 0x7F;
    wrbuf[2] = value;

    *mini_spi_dat.regs.spi_rate_cntl = 3;
    *mini_spi_dat.regs.spi_del_val = 0;

    for ( i = 0; i < 3; i++ )
    {  
        *mini_spi_dat.chip_select.cs_clr[chip_select] = mini_spi_dat.chip_select.mask[chip_select];
        *mini_spi_dat.regs.spi_cfg_reg |= ( 1 << 31 );
        *mini_spi_dat.regs.spi_tx_dat = wrbuf[i];
        while ( ( *mini_spi_dat.regs.spi_int_st & ( 1 << 10 ) ) == 0 );
        val = (*mini_spi_dat.regs.spi_rx_dat) & 0x0FF;
        *mini_spi_dat.chip_select.cs_set[chip_select] = mini_spi_dat.chip_select.mask[chip_select];
        mini_spi_cs_delay();
    }

#if defined( MINI_SPI_USE_MUTEX )
    mutex_unlock( &mini_spi_mutex );
#else
    spin_unlock_irqrestore( &mini_spi_dat.lock, flags );
#endif

#else
	unsigned char wrbuf[4];
	int rval;
	unsigned int count = 0;
	struct spi_message message;
	struct spi_transfer x[3];

	wrbuf[0] = FXO_WR_CMD_CARD(card);
	wrbuf[1] = reg & 0x7f;
	wrbuf[2] = value;

again:
	spi_message_init(&message);
	memset(x, 0, sizeof x);
	x[0].len = 1;
	x[0].tx_buf = wrbuf;
	x[0].cs_change = 1;
	x[1].len = 1;
	x[1].tx_buf = wrbuf+1;
	x[1].cs_change = 1;
	x[2].len = 1;
	x[2].tx_buf = wrbuf+2;
	spi_message_add_tail(&x[0], &message);
	spi_message_add_tail(&x[1], &message);
	spi_message_add_tail(&x[2], &message);

	rval = spi_sync(wc->dev, &message);
	if(rval && count<3)
	{
		count++;
		printk("spi write err %d, %u\n", rval, count);
		goto again;
	}
#endif
}

static unsigned char wctdm_getreg(struct wctdm *wc, int card, unsigned char reg)
{
#if defined( USE_MINI_SPI )
    unsigned char wrbuf[3];
    unsigned char val;
    int i;
    int chip_select;

#if defined( MINI_SPI_USE_MUTEX )
    mutex_lock( &mini_spi_mutex );
#else
    unsigned long flags;
    spin_lock_irqsave( &mini_spi_dat.lock, flags );
#endif
    
    chip_select = wc->dev->chip_select;

    *mini_spi_dat.regs.spi_rate_cntl = 3;
    *mini_spi_dat.regs.spi_del_val = 0;
    wrbuf[0] = FXO_RD_CMD_CARD(card);
    wrbuf[1] = reg & 0x7F;
    wrbuf[2] = 0;

    for ( i = 0; i < 3; i++ )
    {
        *mini_spi_dat.chip_select.cs_clr[chip_select] = mini_spi_dat.chip_select.mask[chip_select];
        *mini_spi_dat.regs.spi_cfg_reg |= ( 1 << 31 );
        *mini_spi_dat.regs.spi_tx_dat = wrbuf[i];
        while ( ( *mini_spi_dat.regs.spi_int_st & ( 1 << 10 ) ) == 0 );
        val = (*mini_spi_dat.regs.spi_rx_dat) & 0x00FF;
        *mini_spi_dat.chip_select.cs_set[chip_select] = mini_spi_dat.chip_select.mask[chip_select];
        mini_spi_cs_delay();
    }

#if defined( MINI_SPI_USE_MUTEX )
    mutex_unlock( &mini_spi_mutex );
#else
    spin_unlock_irqrestore( &mini_spi_dat.lock, flags );
#endif

    return val;
#else
	unsigned char wrbuf[4];
	int rval;
	unsigned int count = 0;
	struct spi_message message;
	struct spi_transfer	x[3];

	wrbuf[0] = FXO_RD_CMD_CARD(card);
	wrbuf[1] = reg & 0x7f;
	wrbuf[2] = 0;
	wrbuf[3] = 0;

again:
	spi_message_init(&message);
	memset(x, 0, sizeof x);
	x[0].len = 1;
	x[0].tx_buf = wrbuf;
	x[0].cs_change = 1;
	x[1].len = 1;
	x[1].tx_buf = wrbuf+1;
	x[1].cs_change = 1;
	x[2].len = 1;
	x[2].rx_buf = wrbuf+2;
	spi_message_add_tail(&x[0], &message);
	spi_message_add_tail(&x[1], &message);
	spi_message_add_tail(&x[2], &message);

	rval = spi_sync(wc->dev, &message);
	if(rval && count<3)
	{
		count++;
		printk("spi read err %d, %u\n", rval, count);
		goto again;
	}

	return wrbuf[2];
#endif
}

#if defined( USE_MINI_SPI )
static void mini_spi_all_cs( int set )
{
    int i;
    struct wctdm *wc;
    uint32_t mask = 0;

	for(i=0;(wc = ifaces[i]);i++)
	{
		if(!wc->check)
            continue;

        if ( set )
            *mini_spi_dat.chip_select.cs_set[wc->dev->chip_select] = mini_spi_dat.chip_select.mask[wc->dev->chip_select];
        else
            *mini_spi_dat.chip_select.cs_clr[wc->dev->chip_select] = mini_spi_dat.chip_select.mask[wc->dev->chip_select];
	}
}

static void gs_read_hook_regs(void)
{
    unsigned char wrbuf[3];
    unsigned char val;
    int i, j;
#if !defined( MINI_SPI_USE_MUTEX )
    unsigned long flags;
#endif
	struct wctdm *wc;
	int card;

    /* Measured 68uS with interrupts disabled */

    for ( i = 0; ( wc = ifaces[i] ); i++ )
    {
        if ( wc->check )
            goto should_check;
    }
    return;

should_check:
#if defined( MINI_SPI_USE_MUTEX )
    mutex_lock( &mini_spi_mutex );
#else
    spin_lock_irqsave( &mini_spi_dat.lock, flags );
#endif

    *mini_spi_dat.regs.spi_rate_cntl = 3;
    *mini_spi_dat.regs.spi_del_val = 0;

    for ( card = 0; card < NUM_CARDS; card++ )
    {
        wrbuf[0] = FXO_RD_CMD_CARD(card);
        wrbuf[1] = 5 & 0x7F;
        wrbuf[2] = 0;

        for ( j = 0; j < 2; j++ )
        {
            mini_spi_all_cs( 0 );
            *mini_spi_dat.regs.spi_cfg_reg |= ( 1 << 31 );
            *mini_spi_dat.regs.spi_tx_dat = wrbuf[j];
            while ( ( *mini_spi_dat.regs.spi_int_st & ( 1 << 10 ) ) == 0 );
            val = *mini_spi_dat.regs.spi_rx_dat;
            mini_spi_all_cs( 1 );
            mini_spi_cs_delay();
        }

        for ( i = 0; ( wc = ifaces[i] ); i++ )
        {
            if ( !(wc->check) )
                continue;

            *mini_spi_dat.chip_select.cs_clr[wc->dev->chip_select] = mini_spi_dat.chip_select.mask[wc->dev->chip_select];
            *mini_spi_dat.regs.spi_cfg_reg |= ( 1 << 31 );
            *mini_spi_dat.regs.spi_tx_dat = wrbuf[2];
            while ( ( *mini_spi_dat.regs.spi_int_st & ( 1 << 10 ) ) == 0 );
            val = (*mini_spi_dat.regs.spi_rx_dat) & 0x00FF;
            wc->reg0shadow[card] = val;
            *mini_spi_dat.chip_select.cs_set[wc->dev->chip_select] = mini_spi_dat.chip_select.mask[wc->dev->chip_select];
            mini_spi_cs_delay();
        }

        wrbuf[0] = FXO_RD_CMD_CARD(card);
        wrbuf[1] = 29 & 0x7F;
        wrbuf[2] = 0;

        for ( j = 0; j < 2; j++ )
        {
            mini_spi_all_cs( 0 );
            *mini_spi_dat.regs.spi_cfg_reg |= ( 1 << 31 );
            *mini_spi_dat.regs.spi_tx_dat = wrbuf[j];
            while ( ( *mini_spi_dat.regs.spi_int_st & ( 1 << 10 ) ) == 0 );
            val = *mini_spi_dat.regs.spi_rx_dat;
            mini_spi_all_cs( 1 );
            mini_spi_cs_delay();
        }

        for ( i = 0; ( wc = ifaces[i] ); i++ )
        {
            if ( !(wc->check) )
                continue;

            *mini_spi_dat.chip_select.cs_clr[wc->dev->chip_select] = mini_spi_dat.chip_select.mask[wc->dev->chip_select];
            *mini_spi_dat.regs.spi_cfg_reg |= ( 1 << 31 );
            *mini_spi_dat.regs.spi_tx_dat = wrbuf[2];
            while ( ( *mini_spi_dat.regs.spi_int_st & ( 1 << 10 ) ) == 0 );
            val = (*mini_spi_dat.regs.spi_rx_dat) & 0x00FF;
            wc->reg1shadow[card] = val;
            *mini_spi_dat.chip_select.cs_set[wc->dev->chip_select] = mini_spi_dat.chip_select.mask[wc->dev->chip_select];
            mini_spi_cs_delay();
        }
    }

#if defined( MINI_SPI_USE_MUTEX )
    mutex_unlock( &mini_spi_mutex );
#else
    spin_unlock_irqrestore( &mini_spi_dat.lock, flags );
#endif
}
#endif

static void gs_check_hook (void)
{
	int i;
	struct wctdm *wc;
	int card;

	for(i=0;(wc = ifaces[i]);i++)
	{
		if(!wc->check) continue;
		for (card=0;card<NUM_CARDS;card++) {
			if (!(wc->cardflag & (1 << card))) continue;
			wctdm_voicedaa_check_hook(wc, card);
		}
	}
	return;
}

static int check_hook_thread(void *unused)
{
	while(!kthread_should_stop())
	{
#if defined( MINI_SPI_BATCH_REGS )
        gs_read_hook_regs();
#endif
		gs_check_hook();
		schedule_timeout_interruptible(check_tv);
	}
	return 0;
}

static inline void wctdm_voicedaa_check_hook(struct wctdm *wc, int card)
{
	unsigned char res;
	long long curkt;
	signed char b;
	int errors = 0;
	struct fxo *fxo = &wc->mod[card].fxo;
	unsigned char lsdrev;
	static long long lastkt = 0;

#if !defined( MINI_SPI_BATCH_REGS )
    wc->reg0shadow[card] = wctdm_getreg(wc, card, 5);
	wc->reg1shadow[card] = wctdm_getreg(wc, card, 29);
#endif

	/* Try to track issues that plague slot one FXO's */
	b = wc->reg0shadow[card];
	if ((b & 0x2) || !(b & 0x8)) {
		/* Not good -- don't look at anything else */
		if (debug)
			printk_ratelimited(KERN_DEBUG "Error (%02x) on FXO%d!\n", b, wc->chans[card]->channo);
		errors++;
	}
	b &= 0x9b;
	if (fxo->offhook) {
		if (b != 0x9)
			wctdm_setreg(wc, card, 5, 0x9);
	} else {
		if (b != 0x8)
			wctdm_setreg(wc, card, 5, 0x8);
	}
	if (errors)
		return;

	curkt = ktime_to_ms(ktime_get());
	if(lastkt)
	{
		if(curkt-lastkt>16) printk("fxo checkd too slow, %lldms, curkt %lld, lastkt %lld\n", curkt-lastkt, curkt, lastkt);
	}
	lastkt = curkt;

	if (!fxo->offhook) {
		if (fwringdetect) {
			res = wc->reg0shadow[card] & 0x60;
			if (fxo->ringdebounce) {
				--fxo->ringdebounce;
				if (res && (res != fxo->lastrdtx) &&
				    (fxo->battery == BATTERY_PRESENT)) {
					if (!fxo->wasringing) {
						fxo->wasringing = 1;
						fxo->ringtimeout = ringtimeout;
						if (debug)
							printk(KERN_DEBUG "RING on FXO%d!\n", wc->chans[card]->channo);
						dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_RING);
					}
					fxo->lastrdtx = res;
					fxo->ringdebounce = 10;
				} else if (!res) {
					if (fxo->ringdebounce ==0 && fxo->wasringing) {
						fxo->wasringing = 0;
						if (debug)
							printk(KERN_DEBUG "NO RING on FXO%d!\n", wc->chans[card]->channo);
						dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
					}
				}
			} else if (res && (fxo->battery == BATTERY_PRESENT)) {
				fxo->lastrdtx = res;
				fxo->ringdebounce = 10;
			}
		} else {
			res = wc->reg0shadow[card];
			if ((res & 0x04) && (fxo->battery == BATTERY_PRESENT)) {
				fxo->ringdebounce += curkt - fxo->ringdebounce_last;
				fxo->ringdebounce_last = curkt;
				if (fxo->ringdebounce >=ringdebounce) {
					if (!fxo->wasringing) {
						fxo->wasringing = 1;
						fxo->ringtimeout = ringtimeout;
						dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_RING);
						if (debug)
							printk(KERN_DEBUG "RING on FXO%d!\n", wc->chans[card]->channo);
					}
					fxo->ringdebounce = ringdebounce;
					fxo->ringdebounce_last = curkt;
				}
			} else {
				fxo->ringdebounce -= curkt-fxo->ringdebounce_last;
				fxo->ringdebounce_last = curkt;
				if (fxo->ringdebounce<=0) {
					if (fxo->wasringing) {
						fxo->wasringing = 0;
						dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
						if (debug)
							printk(KERN_DEBUG "NO RING on FXO%d!\n", wc->chans[card]->channo);
					}
					fxo->ringdebounce = 0;
				}
			}
		}
	}

	b = wc->reg1shadow[card];

	if (fxovoltage) {
		static int count = 0;
		if (!(count++ % 100)) {
			printk(KERN_DEBUG "FXO%d: Voltage: %d Debounce %d\n", wc->chans[card]->channo, b, fxo->battdebounce);
		}
	}

	if (unlikely(DAHDI_RXSIG_INITIAL == wc->chans[card]->rxhooksig)) {
		/*
		 * dahdi-base will set DAHDI_RXSIG_INITIAL after a
		 * DAHDI_STARTUP or DAHDI_CHANCONFIG ioctl so that new events
		 * will be queued on the channel with the current received
		 * hook state.  Channels that use robbed-bit signalling always
		 * report the current received state via the dahdi_rbsbits
		 * call. Since we only call dahdi_hooksig when we've detected
		 * a change to report, let's forget our current state in order
		 * to force us to report it again via dahdi_hooksig.
		 *
		 */
		fxo->battery = BATTERY_UNKNOWN;
	}

	if (abs(b) < battthresh) {
		/* possible existing states:
		   battery lost, no debounce timer
		   battery lost, debounce timer (going to battery present)
		   battery present or unknown, no debounce timer
		   battery present or unknown, debounce timer (going to battery lost)
		*/

		if (fxo->battery == BATTERY_LOST) {
			if (fxo->battdebounce) {
				/* we were going to BATTERY_PRESENT, but battery was lost again,
				   so clear the debounce timer */
				fxo->battdebounce = 0;
			}
		} else {
			if (fxo->battdebounce>0) {
				/* going to BATTERY_LOST, see if we are there yet */
				fxo->battdebounce -= curkt - fxo->battdebounce_last;
				fxo->battdebounce_last = curkt;
				if (fxo->battdebounce<=0) {
					fxo->battery = BATTERY_LOST;
					wc->mod[card].fxo.ringtimeout = 0;
					dahdi_qevent_lock(wc->chans[card], DAHDI_EVENT_BATTERY_LOST);
					if (debug)
						printk(KERN_DEBUG "NO BATTERY on FXO%d!\n", wc->chans[card]->channo);
lsdrev = wctdm_getreg(wc, card, 13);
if(lsdrev!=0x58) printk("unexpect Line-Side Device Revision: 0x%02x\n", lsdrev);
#ifdef	JAPAN
					if (!wc->ohdebounce && wc->offhook) {
						dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_ONHOOK);
						if (debug)
							printk(KERN_DEBUG "Signalled On Hook\n");
#ifdef	ZERO_BATT_RING
						wc->onhook++;
#endif
					}
#else
					dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_ONHOOK);
					/* set the alarm timer, taking into account that part of its time
					   period has already passed while debouncing occurred */
					fxo->battalarm = (battalarm - battdebounce);
					fxo->battalarm_last = curkt;
#endif
				}
			} else {
				/* start the debounce timer to verify that battery has been lost */
				fxo->battdebounce = battdebounce;
				fxo->battdebounce_last = curkt;
			}
		}
	} else {
		/* possible existing states:
		   battery lost or unknown, no debounce timer
		   battery lost or unknown, debounce timer (going to battery present)
		   battery present, no debounce timer
		   battery present, debounce timer (going to battery lost)
		*/

		if (fxo->battery == BATTERY_PRESENT) {
			if (fxo->battdebounce) {
				/* we were going to BATTERY_LOST, but battery appeared again,
				   so clear the debounce timer */
				fxo->battdebounce = 0;
			}
		} else {
			if (fxo->battdebounce>0) {
				/* going to BATTERY_PRESENT, see if we are there yet */
				fxo->battdebounce -= curkt - fxo->battdebounce_last;
				fxo->battdebounce_last = curkt;
				if (fxo->battdebounce<=0) {
					fxo->battery = BATTERY_PRESENT;
					dahdi_qevent_lock(wc->chans[card], DAHDI_EVENT_BATTERY_PRESENT);
					if (debug)
						printk(KERN_DEBUG "BATTERY on FXO%d (%s)!\n", wc->chans[card]->channo,
						       (b < 0) ? "-" : "+");			    
#ifdef	ZERO_BATT_RING
					if (wc->onhook) {
						wc->onhook = 0;
						dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
						if (debug)
							printk(KERN_DEBUG "Signalled Off Hook\n");
					}
#else
					dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
#endif
					/* set the alarm timer, taking into account that part of its time
					   period has already passed while debouncing occurred */
					fxo->battalarm = (battalarm - battdebounce);
					fxo->battalarm_last = curkt;
				}
			} else {
				/* start the debounce timer to verify that battery has appeared */
				fxo->battdebounce = battdebounce;
				fxo->battdebounce_last = curkt;
			}
		}
	}

	if (fxo->lastpol >= 0) {
		if (b < 0) {
			fxo->lastpol = -1;
			fxo->polaritydebounce = POLARITY_DEBOUNCE;
			fxo->polaritydebounce_last = curkt;
		}
	} 
	if (fxo->lastpol <= 0) {
		if (b > 0) {
			fxo->lastpol = 1;
			fxo->polaritydebounce = POLARITY_DEBOUNCE;
			fxo->polaritydebounce_last = curkt;
		}
	}

	if (fxo->battalarm>0) {
		fxo->battalarm -= curkt - fxo->battalarm_last;
		fxo->battalarm_last = curkt;
		if (fxo->battalarm<=0) {
			/* the alarm timer has expired, so update the battery alarm state
			   for this channel */
			dahdi_alarm_channel(wc->chans[card], fxo->battery == BATTERY_LOST ? DAHDI_ALARM_RED : DAHDI_ALARM_NONE);
		}
	}

	if (fxo->polaritydebounce>0) {
		fxo->polaritydebounce -= curkt - fxo->polaritydebounce_last;
		fxo->polaritydebounce_last = curkt;
		if (fxo->polaritydebounce<=0) {
		    if (fxo->lastpol != fxo->polarity) {
				if (debug)
					printk(KERN_DEBUG "FXO%d Polarity reversed (%d -> %d)\n", wc->chans[card]->channo,
				       fxo->polarity, 
				       fxo->lastpol);
				if (fxo->polarity)
					dahdi_qevent_lock(wc->chans[card], DAHDI_EVENT_POLARITY);
				fxo->polarity = fxo->lastpol;
		    }
		}
	}
}

static irqreturn_t fxo_isr(int irq, void *data)
{
	return IRQ_HANDLED;
}

static int wctdm_voicedaa_insane(struct wctdm *wc, int card)
{
	int blah;

//	printk("wctdm_voicedaa_insane\n");

	blah = wctdm_getreg(wc, card, 2);
	if (blah != 0x3)
		return -2;
	blah = wctdm_getreg(wc, card, 11);
	if (debug)
		printk("VoiceDAA System: %02x\n", blah & 0xf);
	return 0;
}

static void wait_just_a_bit(int foo)
{
	long newjiffies;
	newjiffies = jiffies + foo;
	while(jiffies < newjiffies);
}

/*********************************************************************
 * Set the hwgain on the analog modules
 *
 * card = the card position for this module (0-23)
 * gain = gain in dB x10 (e.g. -3.5dB  would be gain=-35)
 * tx = (0 for rx; 1 for tx)
 *
 *******************************************************************/
static int wctdm_set_hwgain(struct wctdm *wc, int card, __s32 gain, __u32 tx)
{
	if (!(wc->modtype[card] == MOD_TYPE_FXO)) {
		printk(KERN_NOTICE "Cannot adjust gain.  Unsupported module type!\n");
		return -1;
	}
	if (tx) {
		if (debug)
			printk(KERN_DEBUG "setting FXO tx gain for card=%d to %d\n", card, gain);
		if (gain >=  -150 && gain <= 0) {
			wctdm_setreg(wc, card, 38, 16 + (gain/-10));
			wctdm_setreg(wc, card, 40, 16 + (-gain%10));
		} else if (gain <= 120 && gain > 0) {
			wctdm_setreg(wc, card, 38, gain/10);
			wctdm_setreg(wc, card, 40, (gain%10));
		} else {
			printk(KERN_INFO "FXO tx gain is out of range (%d)\n", gain);
			return -1;
		}
	} else { /* rx */
		if (debug)
			printk(KERN_DEBUG "setting FXO rx gain for card=%d to %d\n", card, gain);
		if (gain >=  -150 && gain <= 0) {
			wctdm_setreg(wc, card, 39, 16+ (gain/-10));
			wctdm_setreg(wc, card, 41, 16 + (-gain%10));
		} else if (gain <= 120 && gain > 0) {
			wctdm_setreg(wc, card, 39, gain/10);
			wctdm_setreg(wc, card, 41, (gain%10));
		} else {
			printk(KERN_INFO "FXO rx gain is out of range (%d)\n", gain);
			return -1;
		}
	}

	return 0;
}

static int wctdm_init_voicedaa(struct wctdm *wc, int card, int fast, int manual, int sane)
{
	unsigned char reg16=0, reg26=0, reg30=0, reg31=0;
	unsigned int tx_ts, rx_ts;
	long newjiffies;
	wc->modtype[card] = MOD_TYPE_FXO;

//	printk("wctdm_init_voicedaa card = %d, fast = %d, manual = %d, sane = %d\n", card, fast, manual, sane);

	if (!sane && wctdm_voicedaa_insane(wc, card))
		return -2;

	/* Software reset */
	wctdm_setreg(wc, card, 1, 0x80);

	/* Wait just a bit */
	wait_just_a_bit(HZ/10);

	/* enable interrupt */
	wctdm_setreg(wc, card, 4, 0x00);
	wctdm_setreg(wc, card, 2, 0x83);
	wctdm_setreg(wc, card, 3, 0x80);

	/* Enable PCM, ulaw */
	if (alawoverride){
		wctdm_setreg(wc, card, 33, 0x20);
	} else {
		wctdm_setreg(wc, card, 33, 0x28);
	}

	/* Set On-hook speed, Ringer impedence, and ringer threshold */
	reg16 |= (fxo_modes[_opermode].ohs << 6);
	reg16 |= (fxo_modes[_opermode].rz << 1);
	reg16 |= (fxo_modes[_opermode].rt);
	wctdm_setreg(wc, card, 16, reg16);

	/* enable Ring Validation, 14.7Hz-63Hz, timeout 128ms, confirm 100ms, delay 0ms */
	wctdm_setreg(wc, card, 22, 0x0d);
	wctdm_setreg(wc, card, 23, 0x08);
	wctdm_setreg(wc, card, 24, 0x91);

	if(fwringdetect) {
		/* Enable ring detector full-wave rectifier mode */
		wctdm_setreg(wc, card, 18, 2);
	} else { 
		/* Set to the device defaults */
		wctdm_setreg(wc, card, 18, 0);
	}

	/* Set DC Termination:
	   Tip/Ring voltage adjust, minimum operational current, current limitation */
	reg26 |= (fxo_modes[_opermode].dcv << 6);
	reg26 |= (fxo_modes[_opermode].mini << 4);
	reg26 |= (fxo_modes[_opermode].ilim << 1);
	wctdm_setreg(wc, card, 26, reg26);

	/* Set AC Impedence */
	reg30 = (fxo_modes[_opermode].acim);
	if(fxo_override_acim)
	{
		reg30 = fxo_acim;
	}
	wctdm_setreg(wc, card, 30, reg30);

	/* Misc. DAA parameters */
	if (fastpickup)
		reg31 = 0xe3;
	else
		reg31 = 0xa3;

	reg31 |= (fxo_modes[_opermode].ohs2 << 3);
	wctdm_setreg(wc, card, 31, reg31);

	/* Set Transmit/Receive timeslot */
	tx_ts = (wc->iface_index * NUM_CARDS + card) * 8;
	rx_ts = (wc->iface_index * NUM_CARDS + card) * 8;
	wctdm_setreg(wc, card, 34, tx_ts );
	wctdm_setreg(wc, card, 35, tx_ts >> 8 );
	wctdm_setreg(wc, card, 36, rx_ts );
	wctdm_setreg(wc, card, 37, rx_ts >> 8 );

#if 0
	/*16-bit linear*/
	wctdm_setreg(wc, card, 33, 0x38);
#else
	/*8-bit linear*/
	//wctdm_setreg(wc, card, 33, 0x30);
#endif

	/* Enable ISO-Cap */
	wctdm_setreg(wc, card, 6, 0x00);

	if (fastpickup)
		wctdm_setreg(wc, card, 17, wctdm_getreg(wc, card, 17) | 0x20);

	/* Wait 1000ms for ISO-cap to come up */
	newjiffies = jiffies;
	newjiffies += 2 * HZ;
	while((jiffies < newjiffies) && !(wctdm_getreg(wc, card, 11) & 0xf0))
		wait_just_a_bit(HZ/10);

	if (!(wctdm_getreg(wc, card, 11) & 0xf0)) {
		printk("VoiceDAA did not bring up ISO link properly!\n");
		return -1;
	}
	if (debug)
		printk("ISO-Cap is now up, line side: %02x rev %02x\n", 
		       wctdm_getreg(wc, card, 11) >> 4,
		       (wctdm_getreg(wc, card, 13) >> 2) & 0xf);
	/* Enable on-hook line monitor */
	wctdm_setreg(wc, card, 5, 0x08);

	/* Take values for fxotxgain and fxorxgain and apply them to module */
	wctdm_set_hwgain(wc, card, fxotxgain, 1);
	wctdm_set_hwgain(wc, card, fxorxgain, 0);

	/* NZ -- crank the tx gain up by 7 dB */
	if (!strcmp(fxo_modes[_opermode].name, "NEWZEALAND")) {
		printk("Adjusting gain\n");
		wctdm_set_hwgain(wc, card, 7, 1);
	}
	
	if(debug)
		printk("DEBUG fxotxgain:%i.%i fxorxgain:%i.%i\n", (wctdm_getreg(wc, card, 38)/16)?-(wctdm_getreg(wc, card, 38) - 16) : wctdm_getreg(wc, card, 38), (wctdm_getreg(wc, card, 40)/16)? -(wctdm_getreg(wc, card, 40) - 16):wctdm_getreg(wc, card, 40), (wctdm_getreg(wc, card, 39)/16)? -(wctdm_getreg(wc, card, 39) - 16) : wctdm_getreg(wc, card, 39),(wctdm_getreg(wc, card, 41)/16)?-(wctdm_getreg(wc, card, 41) - 16):wctdm_getreg(wc, card, 41));

	return 0;
		
}

static int wctdm_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
	struct wctdm_stats stats;
	struct wctdm_regs regs;
	struct wctdm_regop regop;
	struct wctdm_echo_coefs echoregs;
	struct dahdi_hwgain hwgain;
	struct wctdm *wc = chan->pvt;
	int x;
	switch (cmd) {
	case WCTDM_GET_STATS:
		if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXO) {
			stats.tipvolt = (signed char)wctdm_getreg(wc, chan->chanpos - 1, 29) * 1000;
			stats.ringvolt = (signed char)wctdm_getreg(wc, chan->chanpos - 1, 29) * 1000;
			stats.batvolt = (signed char)wctdm_getreg(wc, chan->chanpos - 1, 29) * 1000;
		} else 
			return -EINVAL;
		if (copy_to_user((__user void *)data, &stats, sizeof(stats)))
			return -EFAULT;
		break;
	case WCTDM_GET_REGS:
		{
			memset(&regs, 0, sizeof(regs));
			for (x=0;x<NUM_FXO_REGS;x++)
				regs.direct[x] = wctdm_getreg(wc, chan->chanpos - 1, x);
		}
		if (copy_to_user((__user void *)data, &regs, sizeof(regs)))
			return -EFAULT;
		break;
	case WCTDM_SET_REG:
		if (copy_from_user(&regop, (__user void *) data, sizeof(regop)))
			return -EFAULT;
		if (regop.indirect) {
			if (wc->modtype[chan->chanpos - 1] != MOD_TYPE_FXS)
				return -EINVAL;
			printk(KERN_INFO "Setting indirect %d to 0x%04x on %d\n", regop.reg, regop.val, chan->chanpos);
		} else {
			regop.val &= 0xff;
			printk(KERN_INFO "Setting direct %d to %04x on %d\n", regop.reg, regop.val, chan->chanpos);
			wctdm_setreg(wc, chan->chanpos - 1, regop.reg, regop.val);
		}
		break;
	case WCTDM_SET_ECHOTUNE:
		printk(KERN_INFO "-- Setting echo registers: \n");
		if (copy_from_user(&echoregs, (__user void *)data, sizeof(echoregs)))
			return -EFAULT;

		if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXO) {
			/* Set the ACIM register */
			wctdm_setreg(wc, chan->chanpos - 1, 30, echoregs.acim);

			/* Set the digital echo canceller registers */
			wctdm_setreg(wc, chan->chanpos - 1, 45, echoregs.coef1);
			wctdm_setreg(wc, chan->chanpos - 1, 46, echoregs.coef2);
			wctdm_setreg(wc, chan->chanpos - 1, 47, echoregs.coef3);
			wctdm_setreg(wc, chan->chanpos - 1, 48, echoregs.coef4);
			wctdm_setreg(wc, chan->chanpos - 1, 49, echoregs.coef5);
			wctdm_setreg(wc, chan->chanpos - 1, 50, echoregs.coef6);
			wctdm_setreg(wc, chan->chanpos - 1, 51, echoregs.coef7);
			wctdm_setreg(wc, chan->chanpos - 1, 52, echoregs.coef8);

			printk(KERN_INFO "-- Set echo registers successfully\n");

			break;
		} else {
			return -EINVAL;

		}
		break;
	case DAHDI_SET_HWGAIN:
		if (copy_from_user(&hwgain, (__user void *) data, sizeof(hwgain)))
			return -EFAULT;

		wctdm_set_hwgain(wc, chan->chanpos-1, hwgain.newgain, hwgain.tx);

		if (debug)
			printk(KERN_DEBUG "Setting hwgain on channel %d to %d for %s direction\n", 
				chan->chanpos-1, hwgain.newgain, hwgain.tx ? "tx" : "rx");
		break;
	default:
		return -ENOTTY;
	}
	return 0;

}

static int wctdm_open(struct dahdi_chan *chan)
{
	struct wctdm *wc = chan->pvt;
	if (!(wc->cardflag & (1 << (chan->chanpos - 1))))
		return -ENODEV;
	if (wc->dead)
		return -ENODEV;
	wc->usecount++;
	if(tdm) tdm->usecount++;

	return 0;
}

static inline struct wctdm *wctdm_from_span(struct dahdi_span *span)
{
	return container_of(span, struct wctdm, span);
}

static int wctdm_close(struct dahdi_chan *chan)
{
	struct wctdm *wc = chan->pvt;
	struct fxs *const fxs = &wc->mod[chan->chanpos - 1].fxs;
	if(tdm) tdm->usecount--;
	wc->usecount--;
	if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXS) {
		int idlehookstate;
		idlehookstate = POLARITY_XOR ?
						SLIC_LF_ACTIVE_REV :
						SLIC_LF_ACTIVE_FWD;
		fxs->idletxhookstate = idlehookstate;
	}
	/* If we're dead, release us now */
	if (!wc->usecount && wc->dead) 
		wctdm_release(wc);
	return 0;
}

static int wctdm_hooksig(struct dahdi_chan *chan, enum dahdi_txsig txsig)
{
	struct wctdm *wc = chan->pvt;
	int chan_entry = chan->chanpos - 1;
	if (wc->modtype[chan_entry] == MOD_TYPE_FXO) {
		/* XXX Enable hooksig for FXO XXX */
		switch(txsig) {
		case DAHDI_TXSIG_START:
		case DAHDI_TXSIG_OFFHOOK:
			wc->mod[chan_entry].fxo.offhook = 1;
			wc->mod[chan_entry].fxo.ringtimeout = 0;
			//wctdm_setreg(wc, chan_entry, 5, 0x9);
			printk("FXO%d offhook\n", chan->channo);
			break;
		case DAHDI_TXSIG_ONHOOK:
			wc->mod[chan_entry].fxo.offhook = 0;
			printk("FXO%d onhook\n", chan->channo);
			//wctdm_setreg(wc, chan_entry, 5, 0x8);
			break;
		default:
			printk(KERN_NOTICE "wcfxo: Can't set tx state to %d\n", txsig);
		}
	}

	return 0;
}

static const struct dahdi_span_ops wctdm_span_ops = {
	.owner = THIS_MODULE,
	.hooksig = wctdm_hooksig,
	.open = wctdm_open,
	.close = wctdm_close,
	.ioctl = wctdm_ioctl,
};

static int wctdm_initialize(struct wctdm *wc)
{
	int x;

	wc->ddev = dahdi_create_device();
	if (!wc->ddev)
		return -ENOMEM;

	/* DAHDI stuff */
	sprintf(wc->span.name, "GXE51XX PSTN Bank");
	snprintf(wc->span.desc, sizeof(wc->span.desc) - 1, "%s Board (%d ports)", wc->variety, wc->pos);
	wc->ddev->location = kasprintf(GFP_KERNEL, "SPI Bus %d.%d ", wc->dev->master->bus_num, wc->dev->chip_select);
	if (!wc->ddev->location) {
		dahdi_free_device(wc->ddev);
		wc->ddev = NULL;
		return -ENOMEM;
	}

	wc->ddev->manufacturer = "Grandstream Networks Inc";
	wc->ddev->devicetype = wc->variety;

	if (alawoverride) {
		printk(KERN_INFO "ALAW override parameter detected.  Device will be operating in ALAW\n");
		wc->span.deflaw = DAHDI_LAW_ALAW;
	} else {
		wc->span.deflaw = DAHDI_LAW_MULAW;
	}
	for (x = 0; x < wc->num_cards; x++) {
		sprintf(wc->chans[x]->name, "GS_FXO/%d/%d", wc->pos, x);
		wc->chans[x]->chanpos = x+1;
		wc->chans[x]->pvt = wc;
	}
	wc->span.chans = wc->chans;
	wc->span.channels = wc->num_cards;
	wc->span.flags = DAHDI_FLAG_RBS;
	wc->span.ops = &wctdm_span_ops;

	list_add_tail(&wc->span.device_node, &wc->ddev->spans);
	if (dahdi_register_device(wc->ddev, &wc->dev->dev)) {
		printk(KERN_NOTICE "Unable to register span with DAHDI\n");
		kfree(wc->ddev->location);
		dahdi_free_device(wc->ddev);
		wc->ddev = NULL;
		return -1;
	}

	return 0;
}

static void wctdm_post_initialize(struct wctdm *wc)
{
	int x;

	/* Finalize signalling  */
	for (x = 0; x < wc->num_cards; x++) {
		if (wc->cardflag & (1 << x)) {
			if (wc->modtype[x] == MOD_TYPE_FXO)
				wc->chans[x]->sigcap = DAHDI_SIG_FXSKS | DAHDI_SIG_FXSLS | DAHDI_SIG_SF | DAHDI_SIG_CLEAR;
			else
				wc->chans[x]->sigcap = DAHDI_SIG_FXOKS | DAHDI_SIG_FXOLS | DAHDI_SIG_FXOGS | DAHDI_SIG_SF | DAHDI_SIG_EM | DAHDI_SIG_CLEAR;
		} else if (!(wc->chans[x]->sigcap & DAHDI_SIG_BROKEN)) {
			wc->chans[x]->sigcap = DAHDI_SIG_FXSKS | DAHDI_SIG_FXSLS | DAHDI_SIG_SF | DAHDI_SIG_CLEAR;;
		}
	}
}

static int wctdm_hardware_init(struct wctdm *wc)
{
	/* Hardware stuff */
	unsigned char x;

	for (x = 0; x < wc->num_cards; x++) {
		int ret=0;
		if (!(ret = wctdm_init_voicedaa(wc, x, 0, 0, 0))) {
			wc->cardflag |= (1 << x);
			printk(KERN_INFO "Module %d: Installed -- AUTO FXO (%s mode)\n",x, fxo_modes[_opermode].name);
		} else
			printk(KERN_NOTICE "Module %d: Not installed\n", x);
	}

	return 0;
}

void si3050_reset(unsigned int gpio)
{
        gpio_direction_output(gpio, 1);
        udelay(100);
        gpio_set_value(gpio, 0);
        udelay(2000);
        udelay(2000);
        udelay(2000);
        udelay(2000);
        gpio_set_value(gpio, 1);
        udelay(2000);
        udelay(2000);
        udelay(2000);
        udelay(2000);
}

//static int gs_wctdm_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
static int gxe51xx_pstn_spi_probe(struct spi_device *spi)
{
	struct si3050_platform_data *pdata = spi->dev.platform_data;
	int ret;
	int res;
	struct wctdm *wc;
	struct wctdm_desc *d = &gs_gxe51xx_desc;
	int x;
	int y;
	int cardcount = 0;

//	printk("gxe51xx_pstn_spi_probe\n");
        if (!pdata) {
                dev_err(&spi->dev, "no platform data provided\n");
                return -ENODEV;
        }

	for (x=0;x<WC_MAX_IFACES;x++)
		if (!ifaces[x]) break;

	if (x >= WC_MAX_IFACES) {
		printk(KERN_NOTICE "Too many interfaces\n");
		return -EIO;
	}

	wc = kmalloc(sizeof(struct wctdm), GFP_KERNEL);
	if (!wc) {
		return -ENOMEM;
	}

	memset(wc, 0, sizeof(struct wctdm));
	wc->num_cards = pdata->chips;
	if(wc->num_cards>NUM_CARDS) wc->num_cards = NUM_CARDS;
	ifaces[x] = wc;
	ifaces_num++;
	wc->pos = x;
	/* Remember which interface are we in the pointer array of interfaces */
	wc->iface_index = x;
	for (x=0; x < sizeof(wc->chans)/sizeof(wc->chans[0]); ++x) {
		wc->chans[x] = &wc->_chans[x];
	}
	spin_lock_init(&wc->lock);
	wc->curcard = -1;
	wc->dev = spi;
	wc->variety = d->name;
	for (y=0;y<wc->num_cards;y++)
		wc->flags[y] = d->flags;

	if (wctdm_initialize(wc)) {
		printk(KERN_NOTICE "wctdm: Unable to intialize FXS\n");
		tdm_remove();
		kfree(wc);
		return -EIO;
	}

	wc->pdata = pdata;
	wc->dev = spi;

	/*
	* setup SPI parameters; this makes sure that parameters we request
	* are acceptable by the spi driver
	*/
	spi_set_drvdata(spi, wc);
	spi->mode = SPI_MODE_0; /* clk active high */
	spi->bits_per_word = 8;
#if defined( USE_MINI_SPI )
    printk("Enter mini spi setup\n");
    mini_spi_setup(spi);
#else
	ret = spi_setup(spi);
	if (ret < 0) {
		dev_err(&spi->dev, "spi_setup() failed\n");
		return -ENODEV;
	}
#endif
	/* Keep track of which device we are */
	spi_set_drvdata(spi, wc);
	dev_info(&spi->dev, "initialized\n");

	wc->irq = MAGIC_INVALID;
	if(MAGIC_INVALID!=wc->pdata->irqno)
	{			
		ret = request_irq(wc->pdata->irqno, fxo_isr, IRQF_DISABLED | IRQF_TRIGGER_FALLING, "fxo_int", wc);
		if(ret < 0) {
			printk("Unable to request IRQ %d\n", wc->pdata->irqno);
		}
		else
		{
			wc->irq = wc->pdata->irqno;
		}
	}

	/* reset FXO chips only the first time. All are controlled by same reset pin */
	if (MAGIC_INVALID!=wc->pdata->reset)
	{
		ret = gpio_request(wc->pdata->reset, "si3050 reset");
		if (ret < 0) {
			dev_err(&spi->dev, "failed requesting reset gpio\n");
			tdm_remove();
			kfree(wc);
			return -EIO;
		}
		si3050_reset(wc->pdata->reset);
	}

	if (wctdm_hardware_init(wc)) {
		spi_set_drvdata(spi, NULL);
		dahdi_unregister_device(wc->ddev);
		kfree(wc->ddev->location);
		dahdi_free_device(wc->ddev);

		kfree(wc);
		/* Release GPIOs */
		if (MAGIC_INVALID!=wc->pdata->reset)
		{
		        gpio_free(wc->pdata->reset);
		}
		return -EIO;
	}

	wctdm_post_initialize(wc);
	for (x = 0; x < wc->num_cards; x++) {
		if (wc->cardflag & (1 << x))
			cardcount++;
	}

	printk("Found a %s (%d modules)\n", wc->variety, cardcount);
	res = 0;
	wc->check = 1;

	return res;
}

static void wctdm_release(struct wctdm *wc)
{
	if(wc->ddev)
	{
		dahdi_unregister_device(wc->ddev);
		kfree(wc->ddev->location);
		dahdi_free_device(wc->ddev);
	}

        /* Remove wc from interface list */ 
	ifaces[wc->iface_index] = NULL;
        ifaces_num--;

	kfree(wc);
	printk(KERN_INFO "Freed a GS PSTN BANK\n");
}



static int __devexit gxe51xx_pstn_spi_remove(struct spi_device *spi)
{
        struct wctdm *wc = spi_get_drvdata(spi);

	printk("gxe51xx_pstn_spi_remove\n");
        if (wc) {

		if(MAGIC_INVALID!=wc->irq) free_irq(wc->irq, wc);
		/* Release GPIOs */
		if (MAGIC_INVALID!=wc->pdata->reset)
		{
			gpio_free(wc->pdata->reset);
		}

                /* Release span, possibly delayed */
                if (!wc->usecount)
                        wctdm_release(wc);
                else
                        wc->dead = 1;
        }

        return 0;
}

static struct spi_driver dmw96_gxe51xx_pstn_spi_driver = {
        .driver = {
                .name   = "dmw96_si3050_spi",
                .bus    = &spi_bus_type,
                .owner  = THIS_MODULE,
        },
        .probe  = gxe51xx_pstn_spi_probe,
        .remove = __exit_p(gxe51xx_pstn_spi_remove),
};

static int __init gs_gxe51xx_pstn_module_init(void)
{
        int x;
	struct sched_param sched_params;

	printk("gs_gxe51xx_pstn_module_init\n");

        for (x = 0; x < (sizeof(fxo_modes) / sizeof(fxo_modes[0])); x++) {
                if (!strcmp(fxo_modes[x].name, opermode))
                        break;
        }
        if (x < sizeof(fxo_modes) / sizeof(fxo_modes[0])) {
                _opermode = x;
        } else {
                printk(KERN_NOTICE "Invalid/unknown operating mode '%s' specified.  Please choose one of:\n", opermode);
                for (x = 0; x < sizeof(fxo_modes) / sizeof(fxo_modes[0]); x++)
                        printk(KERN_INFO "  %s\n", fxo_modes[x].name);
                printk(KERN_INFO "Note this option is CASE SENSITIVE!\n");
                return -ENODEV;
        }

        if (!strcmp(opermode, "AUSTRALIA")) {
                boostringer = 1;
                fxshonormode = 1;
        }

        /* for the voicedaa_check_hook defaults, if the user has not overridden
           them by specifying them as module parameters, then get the values
           from the selected operating mode
        */
        if (battdebounce == 0) {
                battdebounce = fxo_modes[_opermode].battdebounce;
        }
        if (battalarm == 0) {
                battalarm = fxo_modes[_opermode].battalarm;
        }
        if (battthresh == 0) {
                battthresh = fxo_modes[_opermode].battthresh;
        }

	tdm_init(tdm_bus);

	check_hook_task = kthread_run(check_hook_thread, 0, "fxo.checkhookd");
	if(check_hook_task)
	{
		sched_params.sched_priority = check_prio;
		sched_setscheduler(check_hook_task, SCHED_FIFO, &sched_params);
	}

        return spi_register_driver(&dmw96_gxe51xx_pstn_spi_driver);
}
static __devexit void gs_gxe51xx_pstn_module_exit(void)
{
	printk("gs_gxe51xx_pstn_module_exit\n");
	if(check_hook_task)
	{
		kthread_stop(check_hook_task);
		check_hook_task = NULL;
	}
	tdm_remove();
        spi_unregister_driver(&dmw96_gxe51xx_pstn_spi_driver);
}

module_param(disable_led, uint, 0600);
module_param(ringtimeout, int, 0600);
module_param(fxo_acim, uint, 0600);
module_param(fxo_override_acim, uint, 0600);
module_param(check_prio, uint, 0600);
module_param(check_tv, uint, 0600);
module_param(tdm_dma_chunks, uint, 0600);
module_param(tdm_bus, uint, 0600);
module_param(tdm_srate, uint, 0600);
module_param(tdm_slots, uint, 0600);
module_param(tdm_sbits, uint, 0600);
module_param(tdm_rxedge, uint, 0600);
module_param(tdm_txedge, uint, 0600);
module_param(tdm_master, uint, 0600);
module_param(tdm_dma, uint, 0600);

module_param(debug, int, 0600);
module_param(fxovoltage, int, 0600);
module_param(loopcurrent, int, 0600);
module_param(reversepolarity, int, 0600);
module_param(robust, int, 0600);
module_param(opermode, charp, 0600);
module_param(timingonly, int, 0600);
module_param(lowpower, int, 0600);
module_param(boostringer, int, 0600);
module_param(fastringer, int, 0600);
module_param(fxshonormode, int, 0600);
module_param(battdebounce, uint, 0600);
module_param(battalarm, uint, 0600);
module_param(battthresh, uint, 0600);
module_param(ringdebounce, int, 0600);
module_param(dialdebounce, int, 0600);
module_param(fwringdetect, int, 0600);
module_param(alawoverride, int, 0600);
module_param(fastpickup, int, 0600);
module_param(fxotxgain, int, 0600);
module_param(fxorxgain, int, 0600);
module_param(fxstxgain, int, 0600);
module_param(fxsrxgain, int, 0600);

MODULE_DESCRIPTION("GS Asterisk Appliance 51XX PSTN bank Driver");
MODULE_ALIAS("gxe51xx_pstn");
MODULE_LICENSE("GPL");

module_init(gs_gxe51xx_pstn_module_init);
module_exit(gs_gxe51xx_pstn_module_exit);
